/****************************************************************************
** sprtalay.dll V0.2
**
** Diese DLL ist fr MegaCAD evo.II geschrieben. 
** Die DLL speichert alle Layer der aktuell geffneten Zeichnung getrennt 
** nach Layern, unter angabe des aktuellen Zeichungsnamen, der Layernummer 
** und ihres Layernamens in das aktuelle Verzeichnis. 
** Leider funktioniert das speichern nur 2D, da MegaCAD nur eine 2D Speicherung
** einzelner Elemente von aussen ber die C-Schnittstelle zugnglich macht.
** NEU:
** Alle Funktionsaufrufe sind in MegaCAD oder WinAPI enthalten, es ist keine 
** QT-DLL mehr erforderlich.
** Fehlerhafte Typen und Werte wurden berichtigt.
** Eine Fortschrittsanzeige wurde in die Statuszeile von MegaCAD integiert.
** Die Speicherung ist mittels ESC-Taste abbrechbar.
** Sonderzeichen im Layername wie z.B. \|/:*?"<> werden jetzt in einem 
** Bindestich '-' umgewandelt, um einen gltigen Dateinamen zu erhalten.
**
** Created: Sat Okt 04 10:42:06 2003
** 
** by Karsten Hiller 
** 
****************************************************************************/

#include <StdAfx.h>
#include <lg_error.h>
#include <fileNameFunctions.h>

#define ELEMENT_RANGE 1000		// Anfangsgrsse und Schrittweite fr Element Array
#define NAME_MAXLEN _MAX_FNAME	// Lnge der Dateinamen 255 o. 512 Zeichen (WIN9x/WIN2K)

#define PROGTICK 50				// Zeitinterval fr Progress-Aktualisierung in ms
								// min NT>3.5=10ms, NT>3.1=16ms, WIN>95=55ms

typedef struct _array
{ ulong	  ulSize;				// Grsse des Arrays
  ulong*  pFirst;				// Zeiger auf erstes Element des Arrays
} t_array;

//globale Variablen, knnten auch als definierte struktur SelElmLay bergeben werden,
//aber so ist das ganze nachvollziebarer
//t_entity	  elmpara			// als erstes in struct fr megacad notwendig
t_array		  elmList;			// Array fr Elementliste (ulong)
ulong		  elmPos = 0;		// Zhler fr die akt. Position in elmList
uchar		  elmData[32000];	// BytePuffer fr variable ElementDaten (z.Z.max.32000)  
ushort		  elmLayNum = 0;	// LayerNummer der auszuwhlenden Elemente


/*****************************************************************************
** Reserviert Speicher fr ein dynamisches Array vom Typ ulong
** das Array kann nur erzeugt, vergrssert und gelscht werden	
** *currentArray - Zeiger auf Array Structur vom Typ t-array
** ulRange - Wert (Grsse) um die das Array vergrssert wird, ist der Wert 
**			 von ulRange 0 wird das Array aus dem Speicher entfernt
** der Rckgabewert ist TRUE wenn das Array erfolgreich erzeugt oder ver-
** grssert werden konnte, falls es ein Problem gab ist der Rckgabewert FALSE			  	
*****************************************************************************/
bool Array(t_array *currentArray, ulong ulRange)
{
  t_array newArray;
  if( ulRange != 0 )
  {
	if( currentArray->ulSize == 0 ) //neudefinieren falls altelnge = 0
	{
	  currentArray->pFirst = new ulong [ ulRange ];
	  if( currentArray->pFirst == NULL )
	  {
		return( FALSE ); //Fehler
	  }
	  else
	  {
		currentArray->ulSize = ulRange;	//Anfangs Gre
	  }
	}
	else
	{
	  newArray.ulSize  = (currentArray->ulSize + ulRange); // neue Grsse
	  newArray.pFirst = new ulong [ newArray.ulSize ];
	  if( newArray.pFirst == NULL ) 
	  {
		return( FALSE ); //Fehler
	  }
	  else
	  {
		for( ulong i = 0; i < currentArray->ulSize; i++) // Inhalte kopieren
		{
		  newArray.pFirst[i] = currentArray->pFirst[i];
		}
		delete[] currentArray->pFirst;			// alten Speicherbereich freigeben
		currentArray->pFirst = newArray.pFirst;	// neuen Speicherbereich zuweisen
		currentArray->ulSize  = newArray.ulSize;// neue Grsse zuweisen
	  }
	}
  }
  else
  {
	delete[] currentArray->pFirst;	 // Speicherbereich freigeben
	currentArray->ulSize = 0;		 // Grsse auf Null setzen
  }
  return( TRUE );
}

/*****************************************************************************
** berprfung auf Unterbrechung (ESC-Taste)
** gibt TRUE zurck wenn die ESC-Taste gedrckt wurde sonst FALSE				  	
*****************************************************************************/
bool Interrupt()
{
  static bool interrupt = FALSE;  //Variable fr Abbruch dauerhaft merken
  MSG lpMsg;					  //Variable fr Nachrichten Strucktur

  //prft auf Tastaturnachichten, rumt diese in lpMsg ein und entfernt die 
  //die Nachricht aus der ThreadMessageQueue (WinAPI)
  PeekMessage( &lpMsg, NULL, WM_KEYFIRST, WM_KEYLAST, PM_REMOVE); 
  
  //falls eine Taste gedrckt ist und dies die ESC-Taste ist, dann ...
  if( lpMsg.message == WM_KEYDOWN && VK_ESCAPE == (int)lpMsg.wParam )
  {
	if( interrupt == FALSE )  //ist das Programm schon vorher abgebrochen worden
	{
	  //Programmabbruch besttigen und entsprechend der Antwort interrupt setzen
	  switch( MessageBox( NULL, "Programm wirklich abbrechen?",
								"Programm Abbruch", 
								 MB_ICONQUESTION | MB_YESNO ) )
	  {
		case IDNO   :  interrupt = FALSE;  break;
		case IDYES  :  interrupt = TRUE;   break;
		default		:  interrupt = FALSE;  break;
	  }
	}
  }
  return( interrupt );
}


/*****************************************************************************
** Hilfe Fenster
** Werte siehe MegaCAD-C Funktionen 				  	
*****************************************************************************/
short HelpWindow(void* parg)
{
  char text[] = "sprtalay.dll - dient dazu alle in einer Zeichnung vorhanden Layer "
				"getrennt nach ihren Namen zu speichern.\n\n"
				"Die Speicherung erfolgt dabei unter Verwendung des aktuellen Pfades "
				"und Zeichnungsnamens, welche in einem Dialog gendert werden knnen.\n"
				"Zwecks Unterscheidung der unterschiedlichen Layer wird an den "
				"Zeichnungsnamen jeweils getrennt durch einen Unterstich die Layernummer und "
				"der Layername angehngt. Sonderzeichen in Layernamen wie z.B. \\|/\":*?.<> "
				"werden durch eine Bindestich ersetzt, damit ein gltiger Dateiname erzeugt "
				"werden kann.\n"
				"Die Angabe der Layernummer erfolgt aus Sicherheitsgrnden, damit "
				"gleichnamige Layer nicht einfach berschrieben werden.\n"
				"Leere Layer d.h. Layer welche zwar einen Namen besitzen, aber keine "
				"Elemente werden nicht gespeichert.\n\n"
				"Leider knnen alle Daten nur in 2D gespeichert werden, da MegaCAD nur eine "
				"2D-Funktion zum speichern einzelner Elementdaten bereitstellt.\n\n "
				"by K.Hiller (2003.10.04) Version 0.2\n\n"
				"This is FREEWARE \0";
  MessageBox(NULL, text, "sprtalay.dll help window", MB_ICONINFORMATION | MB_OK );
  return( TRUE );
}

/*****************************************************************************
** Progress Fortschrittsanzeige in der Statuszeile von MegaCAD
** ulong pos  - Wert fr aktuellen Fortschritt (xx%)
** ulong max  - maximaler Wert fr aktuellen Fortschritt (100%)
** char* text -	Zustzlicher Anzeigetext			  	
*****************************************************************************/
void Progress(ulong pos=0, ulong max=0, char* text=NULL)
{
  static DWORD  progTime[2]	  = {0,0};		  //Prozesstime fr Elemente
  static ushort progCount	  = 0;			  //Progresszhler
  static char   progress[11]  = "__________"; //LayerFortschritt
  char spin[2]		= "-";					  //Bearbeitungsanzeige 
  char buffer[100]	= "\0";					  //Zeichenbuffer fr Ausgabe
  char defaulttext[]= "Speicherung der Zeichnung getrennt nach Layern";

  progTime[1] = (DWORD)GetTickCount()/PROGTICK;
  if( progTime[0] != progTime[1] )
  {
	progTime[0] = progTime[1]; //Zeit merken
	++progCount;
	switch( progCount )	  //Bearbeitungsanzeige setzen
	{
	  case  1: spin[0] = '\\'; break;
	  case  2: spin[0] = '|' ; break;
	  case  3: spin[0] = '/' ; break;
	  default: spin[0] = '-' ; progCount=0; break;
	}

	if( pos != 0 && max != 0 && pos <= max )  //Fortschritt neu setzen, wenn ...
	{
	  //LayerFortschritt mit * auffllen (maximal 10 Stck)
	  memset( progress, 42, (int)((10.0f/max*pos)+0.51f) );
	}

	if( (text == NULL) || (*text == '\0') )	  //falls kein Text default Text
	{
	  text = defaulttext;
	}
	//Text fr Fortschrittsanzeige zusammen stellen
	sprintf( buffer, "%s %s %s", spin, progress, text );	
	SetFuncText( buffer );		//Statuszeile setzen
  }	
  return;
}


/*****************************************************************************
** Selektiert alle Elemente eines bestimmten Layers und					  	
** trgt dessen Element ID in ein dynamisches Array ein						
** dessen Grsse ggf. vergrssert wird.	
** Werte siehe MegaCAD-C Funktionen 				  	
*****************************************************************************/
short SelElmLay( void* parg )
{
  char		buffer[200];  //Zwischenspeicher fr text
  t_entity* ent;		  //Zeiger auf Elementstrucktur

  ent = (t_entity*) parg; //Addresse auf t_entity aus bergebenen untypisierten Parameter
						  //mittels typecast lokaler Zeigervariabel 'ent' zuweisen
  
  if( TRUE == Interrupt() ) return( BREAK_LOOP ); //auf Unterbrechung prfen
  
  if( elmPos == elmList.ulSize )  //Teste ob Speicherplatz in elmList noch reicht
  {						
	//vergrssert die elmList um ELEMENT_RANGE, gibt eine Fehlermeldung aus
	//falls dies nicht mglich war und bricht die ElementSchleife 'SelectLoop' ab
	if( FALSE == Array( &elmList, ELEMENT_RANGE ) ) 
	{	  
	  sprintf( buffer, "Es konnte kein zustzlicher Speicher reserviert werden.\n"
					   "Es werden Daten fehlen!\n"
					   "ArraySize: %i\n"
					   "Element: %i", elmList.ulSize, elmPos);    
	  MessageBox( NULL, buffer, "Function - SelElmLay", MB_ICONWARNING | MB_OK );
	  return( BREAK_LOOP );	//zurck zur ElementSchleife und Abbruch d. Schleife
	}
  }
  if( ent->attr.lay == elmLayNum )	  //Vergleiche Layernummern
  {
	elmList.pFirst[elmPos] = ent->id; //speichert die Elementnummer in elmListe
	++elmPos;						  //erhht Zhler de elmListe um 1
	Progress();						  //Fortschrittsanzeige in Statuszeile, Spinn
  }
  return( CONTINUE_LOOP );			  //zurck zur ElementSchleife ohne Abbruch
}


/*****************************************************************************
**
** Haupprogamm
** mit Schleife ber alle Layer und Speicherung der ElementDaten als 2D-PRT
**
*****************************************************************************/
short MegaMain( char* file,
				char* args )
{
  h_LGList	  hLayLGList	= NULL;				//Handle der Layerliste
  h_LGItem	  layItem		= NULL;				//Handle des LayerItem
  ulong		  layAnzahl		= 0;				//Menge der Layer	
  ushort	  layErr		= 0;				//ErrorCodeVariable
  t_LayGrpDta layData;							//Struktur der Layerdaten
  char		  saveName[NAME_MAXLEN]		= "\0";	//Name der zuspeichernden Datei
  char		  saveNameLay[NAME_MAXLEN]	= "\0";	//Name der zuspeichernden Datei + Layername
  char		  validFileName[NAME_MAXLEN]= "\0";
  char		  filePathComp[NAME_MAXLEN]	= "\0";	//gekrzter Name der zuspeichernden Datei
  char		  buffer[NAME_MAXLEN]		= "\0"; //buffer fr testausgabe

  t_entity ent;						//GebildeStrucktur
  t_mat44  ecs;						//3D-MatrixStrucktur
  
  ulong		filter1 = 0xffffffff;	// Selektionskriterium -> (2D) alle Elemente erlaubt
  ulong		filter2 = 0xffffffff;	// Selektionskriterium -> (3D) alle Elemente erlaubt
  

  Array( &elmList, ELEMENT_RANGE );	//Speicherplatz fr elmList einrumen

  // Text in der Statuszeile von MegaCAD
  SetFuncText( "Speicherung der aktuellen Zeichnung getrennt nach Layern" );
  // Hilfe Funktion der F1-Taste zuweisen (callback function)
  SetHelpFunc( HelpWindow );              

  //Name der geladenen Zeichnung ermitteln inc. Typumwandlung 
  //wenn eine Zeichnung geladen ist diese Layerweise Speichern currentName
  if( 0 != GetCurrentFile( reinterpret_cast<uchar*>(saveName) ) )		
  {												
	// let the user select a filename, if OK then then do it
	if( TRUE == GetFilename( "Speichern aller Layer in einzelnen Dateien", saveName ) )
	{
	  //Filename am Punkt abschneiden (.prt)
	  strtok( saveName, "." );			
	
	  //erzeuge Kopie/Handle der Layerliste
	  hLayLGList = LG_CreateList(LGL_LAYER, &layErr);	
	  LG_Error("LGL-LAYER-CREATE", &layErr);			//Fehlerbehandlung mit Message
	
	  //ermittelt die Anzahl der Layer
	  layAnzahl = LG_GetInfo(hLayLGList, LGINF_GETCOUNT, &layErr);
	  LG_Error("LGL-LAYER-GETINFO", &layErr);			//Fehlerbehandlung mit Message
	
	  layItem = LGITEM_ROOT;							//setze auf ROOT-item (Anfang)
	  for( ushort i=0; i<=layAnzahl; i++)				//Schleife ber alle Layer
	  {
		//holt immer den linear nchsten Layer aus der LayerStrucktur
		layItem = LG_ItemOp(hLayLGList, layItem, 0, &layData, LGOP_LINEARNEXT, &layErr);
		//Fehlerbehandlung mit Message, bei Fehler Schleife abbrechen
		if( IDABORT == LG_Error("LGL-LAYER-ITEM", &layErr) ) break;
		//auf Unterbrechung prfen
		if( TRUE == Interrupt() ) break;

		elmLayNum = layData.num;			 //Layernummer ermitteln
		
		elmPos = 0;							 //Elementzhler auf Null setzen
		// Schleife ber alle Elemente 
		EntityLoop( filter1,filter2,SelElmLay,
	  				&ent,&elmData,&ecs,sizeof elmData );
	
		saveNameLay[0] = '\0';
		if( elmPos != 0 )	//teste ob berhaupt Elemente im Layer wahren 
		{
		  //erzeugt einen gltigen Dateinamen falls layData.txt Sonderzeichen enthlt
		  MakeValidFileName( validFileName, layData.txt );
		  //Speichernamen zusammenstellen (Zeichungsnamen_Layernummer_Layername.prt)
		  sprintf( saveNameLay, "%s_%u_%s.prt\0", saveName, elmLayNum, validFileName);
		  // save entities to 2D.PRT, Elemente des Layers speichern
		  // bei einem Fehler Message anzeigen
		  if( FALSE == SaveToPrt( saveNameLay, elmList.pFirst, elmPos, 0L) )
		  {
			sprintf( buffer, "Nichts gespeichert. \nFile: %s", saveNameLay );
			MessageBox(NULL, buffer, "Fehler beim speichern", MB_ICONWARNING | MB_OK );
		  }
		}
		//Progressbar in Statuszeile
		Progress( elmLayNum, layAnzahl, FilePathComp( filePathComp, saveNameLay, 40) );	 
	  }
	}
  }
  else
  {
	MessageBox(NULL, "Keine Zeichung geladen.\n"
					 "Es knnen keine Daten gespeichert werden!", 
					 "Fehler beim ffnen", MB_ICONWARNING | MB_OK);
  } 
  LG_Replace(hLayLGList, NULL, 1);	  //Handle (1) fr Layer freigeben
  Array( &elmList, 0 );				  //Speicherplatz fr elmList freigeben
  return TRUE;
}

